// generation.cpp: implementation of the generation class.
//
//////////////////////////////////////////////////////////////////////

//#include "stdafx.h"
#include "generation.h"
#include <stdlib.h>
#include <fstream.h>
#include <iostream.h>
#include <time.h>

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

Generation::Generation(Log *log, int generation)
{
	this->log=log;
	log->addArea((long)this,"Generation no",generation);
	srand((unsigned)time(NULL));
	this->generation=generation;
	for (int i=0; i<POPULATION_SIZE; ++i) {
		kickTree[i] = NULL; //new DecisionTree(log,KICK,rand());
		moveTree[i] = NULL; //new DecisionTree(log,MOVE,rand());
	}
	stats=false;
}

Generation::~Generation()
{
	for (int i=0; i<POPULATION_SIZE; ++i) {
		delete kickTree[i];
		delete moveTree[i];
	}
}

void Generation::createNewPopulation(int minKickTreeSize, int maxKickTreeSize, int minMoveTreeSize, int maxMoveTreeSize)
{
  log->log((long)this,"entering createNewPopulation");
  for (int i=0; i<POPULATION_SIZE; ++i) {
    kickTree[i] = new DecisionTree(log,KICK,rand());
    moveTree[i] = new DecisionTree(log,MOVE,rand());
    kickTree[i]->createRandomSubTree(1,minKickTreeSize,maxKickTreeSize,0,PREDICATE);
    moveTree[i]->createRandomSubTree(1,minMoveTreeSize,maxMoveTreeSize,0,PREDICATE);  
  }
  generation = 0;
  log->log((long)this,"leaving createNewPopulation");
}

void Generation::print(char *filename)
{
	log->log((long)this,"Entering print(\"",filename,"\")...");
	log->debug((long)this,"gen=",this->generation);
	ofstream  f;
	f.open(filename,ios::out|ios::trunc);
	f<<"######################################\n";
	f<<"## GENERATION "<<this->generation<<endl;
	f<<"######################################\n\n";
	for (int i=0; i<POPULATION_SIZE; ++i) {
		f<<"=====================================\n";
		f<<"Invidiual #"<<i+1;
		if (stats)
			f<<" (rank="<<rank[i]<<",fitness="<<fitness[i]<<")\n";
		f<<endl;
		f<<"=====================================\n";
		f<<"---------\nKick Tree\n---------\n";
		kickTree[i]->print(&f);
		f<<"---------\nMove Tree\n---------\n";
		moveTree[i]->print(&f);
		f<<"\n\n";
	}
	f.close();
	log->log((long)this,"Leaving print(\"",filename,"\")");
}

void Generation::loadStats(const char *filename)
{
  char tmpstr[50];
  strcpy(tmpstr,filename);
  log->log((long)this,"Entering loadStats(\"",tmpstr,"\")...");
  int i, k;
  bool ok;
  // load file
  ifstream f;
  f.open(tmpstr,ios::in|ios::binary);
  for (i=0; i<POPULATION_SIZE; ++i) {
    f.read((char *)&fitness[i], sizeof(float));
    inv_rank[i]=i;
  }
  f.close();
  // determine rank
  do {
    ok = true;
    for (i=0; i<POPULATION_SIZE-1; ++i)
      if (fitness[inv_rank[i]]<fitness[inv_rank[i+1]]) {
	k = inv_rank[i];
	inv_rank[i] = inv_rank[i+1];
	inv_rank[i+1] = k;
	ok = false;
      }
  } while (!ok);
  for (i=POPULATION_SIZE-1; i>=0; --i) {
    rank[inv_rank[i]]=i+1;
    inv_rank[i+1]=inv_rank[i];
  }
  stats=true;
  log->log((long)this,"Leaving loadStats(\"",tmpstr,"\")...");
}

Generation* Generation::evolve()
{
	log->log((long)this,"Entering evolve()...");
	Generation* next = new Generation(this->log,this->generation+1);
	log->debug(1," PM = ",(int)(100.0*PM));
	int i=0, j=0, k=0;
	int x, y;
	float Pc = PC/(2-PC);
	DecisionTree **d;
	while (i<POPULATION_SIZE) {
		x = select();
		y = select();
		//crossover
		if (i+1<POPULATION_SIZE && randP(Pc)) {
			d = this->moveTree[x]->crossover(this->moveTree[y]);
			next->moveTree[i]   = d[0];
			next->moveTree[i+1] = d[1];
			d = this->kickTree[x]->crossover(this->kickTree[y]);
			next->kickTree[i]   = d[0];
			next->kickTree[i+1] = d[1];
			i+=2;
		}
		//reproduce and eventually mutate
		else {
			next->moveTree[i] = this->moveTree[x]->clone();
			next->kickTree[i] = this->kickTree[x]->clone();
			if (randP(PM/(1-PC))) {
			  next->moveTree[i]->mutate();
			  next->kickTree[i]->mutate();
			  k++;
			}
			i++;
			j++;
		}
	}
	log->log((long)this,"#inv. reproduced = ",j-k);
	log->log((long)this,"#inv. mutated = ",k);
	log->log((long)this,"Leaving evolve()");
	return next;
}

int Generation::select()
{
  float p[POPULATION_SIZE+1], s;
  int i, n=POPULATION_SIZE;
  p[0]=0;
  
  //Old linear eq.
  //for (i=1; i<=n; ++i)
  //  p[i] = p[i-1] + 1 + (QUOTA-1)*(float)(n-i)/(n-1);
  
  //New quad. eq.
  float q=((float)QUOTA/(float)(n*n)-1.0)/(1.0-(float)QUOTA);
  for (i=1; i<=n; ++i)
    p[i] = p[i-1] + ((n+1-i)/(float)n)*((n+1-i)/(float)n) + q;
  

  s = (float)rand()*p[n]/(float)0x80000000;
  i = 1;
  while (s > p[i])
    ++i;
  if (i<1 || i>POPULATION_SIZE)
    log->error((long)this,"select()");
  return inv_rank[i];
}

void Generation::save(char *filename)
{
	log->log((long)this,"Entering save(\"",filename,"\")...");
	int i;
	ofstream f;
	f.open(filename, ios::out|ios::trunc|ios::binary);
	f.write((char *)&POPULATION_SIZE,sizeof(int));
	f.write((char *)&generation,sizeof(int));
	f.write((char *)&stats,sizeof(bool));
	for (i=0; i<POPULATION_SIZE; ++i)
		f.write((char *)&fitness[i],sizeof(float));
	for (i=0; i<POPULATION_SIZE; ++i)
		f.write((char *)&rank[i],sizeof(int));
	for (i=1; i<=POPULATION_SIZE; ++i)
		f.write((char *)&inv_rank[i],sizeof(int));
	for (i=0; i<POPULATION_SIZE; ++i) {
		moveTree[i]->save(&f);
		kickTree[i]->save(&f);
	}
	f.close();
	log->log((long)this,"Leaving save(\"",filename,"\")");
}

bool Generation::load(char *filename)
{
	log->log((long)this,"Entering load(\"",filename,"\")...");
	int i;
	for (i=0; i<POPULATION_SIZE; ++i) {
		if (kickTree[i] == NULL)
			kickTree[i] = new DecisionTree(log,KICK,rand());
		if (moveTree[i] == NULL)
			moveTree[i] = new DecisionTree(log,MOVE,rand());
	}
	ifstream f;
	f.open(filename,ios::in|ios::binary);
	f.read((char *)&i,sizeof(int)); 
	if (f.eof()) {
		log->log((long)this,"Leaving load(\"",filename,"\") - file does not exists");
		return false;
	}
	if (i!=POPULATION_SIZE) 
		log->error((long)this,"load() - POPULATION_SIZE");
	f.read((char *)&generation,sizeof(int));
	f.read((char *)&stats,sizeof(bool));
	for (i=0; i<POPULATION_SIZE; ++i)
		f.read((char *)&fitness[i],sizeof(float));
	for (i=0; i<POPULATION_SIZE; ++i)
		f.read((char *)&rank[i],sizeof(int));
	for (i=1; i<=POPULATION_SIZE; ++i)
		f.read((char *)&inv_rank[i],sizeof(int));
	for (i=0; i<POPULATION_SIZE; ++i) {
		moveTree[i]->load(&f);
		kickTree[i]->load(&f);
	}
	if (f.eof())
		log->error((long)this,"load() - read error");
	log->addArea((long)this,"Generation no",generation);
	f.close();
	log->log((long)this,"Leaving load(\"",filename,"\")");
	return true;
}

void Generation::reset()
{
	this->stats = false;
	this->generation = 0;
	for (int i=0; i<POPULATION_SIZE; ++i) {
		moveTree[i]->reset();
		kickTree[i]->reset();
	}
}

bool Generation::randP(float p)
{
        unsigned int upper = 0x7FFFFFFF;
	//log->debug(1," (int)upper = ",(int)upper);
	if (p<0.0 || p>1.0)
		log->error((long)this,"randP(...)");
	int th = (int)(p*(float)upper);
	if (th<0)
	  return true;
	if (rand()<=th)
		return true;
	return false;
}
